import os
from ruamel.yaml import YAML
import subprocess

HOME_CONFIG_FILE=os.path.join(os.environ.get('HOME'), '.config/unsubmail/config.yaml')
SERVER = 'server'
USER = 'user'
PASSWORD = 'password'
MAILBOXES= 'mailboxes'

def depass(password):
    if password and password.startswith('pass:'):
        key = password[5:]
        
        print(f"Reading password from key: {key}")
        result = subprocess.run(['pass', key], stdout=subprocess.PIPE)
        return result.stdout.decode('utf-8')
    
    return password


def load(config_file=None):
    config_path = config_file or HOME_CONFIG_FILE

    if os.path.exists(config_path) is not True:
    	raise Exception(f"No configuration defined in {config_file} or {HOME_CONFIG_FILE}")
    	
    data = {}

    print(f"Loading config from: {config_path}")
    with open(config_path) as f:
        data = YAML().load(f)

    errors = []
    for key in (SERVER, USER, PASSWORD):
        if data.get(key) is None:
            errors.append(key)

    if len(errors) > 0:
        raise Exception(f"You must provide configuration for {errors} in {config_path}")

    data[PASSWORD] = depass(data[PASSWORD])

    return Config(data)

class Config:
    def __init__(self, data):
        self.server = data.get(SERVER)
        self.user = data.get(USER)
        self.password = data.get(PASSWORD)
        self.mailboxes = data.get(MAILBOXES)


import requests
from urllib.parse import urlparse
from bs4 import BeautifulSoup as bs
import json
import re

def ngpvan_unsub(url, baseurl, html, form_obj, form_params):
    match = re.search(r"\$scope.contact\s*=\s*(\{[^;]+);", html)
    if match is not None:
        contact = json.loads(match.group(1))

        form_params.update({
            'id': contact.get('VanId'),
            'emailMessageContentId': contact.get('EmailMessageContentId'),
            'emailAddress': contact.get('EmailAddress'),
            'unsubscribe': True,
            'distributionId': contact.get('DistributionId'),
            'distributionUniqueId': contact.get('DistributionUniqueId')
        })

        try:
            resp = requests.post(f"{baseurl}/Unsubscribe/Update", params=form_params)
            print(f"ngpvan Unsubscribe response: {resp.status_code}")
            return resp.status_code == 200
        except Exception as e:
            print(f"Failed to send HTTP request for unsubscribe: {str(e)}")

        return False

def myngp_unsub(url, baseurl, html, form_obj, form_params):
    form_params['DoNotEmail']=True
    try:
        resp = requests.post(url, params=form_params)
        print(f"myngp Unsubscribe response: {resp.status_code}")
        return resp.status_code == 200
    except Exception as e:
        print(f"Failed to send HTTP request for unsubscribe: {str(e)}")

    return False


fnmap = {
    'email.ngpvan.com': ngpvan_unsub,
    'act.myngp.com': myngp_unsub
}

def process_unsub(url):
    print(f"Processing:\n\n    {url}\n\n")

    try:
        resp = requests.get(url)
    except Exception as e:
        print(f"Failed to send initial HTTP request for unsubscribe: {str(e)}")
        return False

    # print(resp)

    if resp.status_code != 200:
        print(f"The unsubscribe request appears to have failed! {resp.status_code}")

    soup = bs(resp.text, 'html.parser')

    up = urlparse(url)
    urlpath = up.path
    # print(f"\nOriginal URL path: '{urlpath}'")
    idx = url.index(urlpath)

    baseurl = url[0:idx]
    print(f"Got base URL: {baseurl}")

    if soup.body is None:
        return True

    forms = soup.body.find_all('form')
    form = None
    for f in forms:
        if 'search' not in str(f):
            form = f
            break

    if form is not None:
        # print(f"Processing form: {form}")

        valmap = {}
        for ipt in form.find_all('input'):
            if ipt['type'] in ('button', 'submit') or ipt.get('name') is None:
                continue

            valmap[ipt['name']] = ipt.get('value') or ''


        fn = fnmap.get(up.hostname)
        if fn is not None:
            print(f"Passing on to unsubscribe function: {fn}")
            return fn(url, baseurl, resp.text, form, valmap)

        else:
            print(f"Pulling params out of form:\n\n    {form.attrs}\n")
            submit = form.get('action')
            print(f"Raw form submit action: '{submit}'")

            if submit is None:
                print("Cannot automatically submit unsubscribe form. Skipping")
                return False

            elif len(submit) < 1:
                submit = url
            
            elif submit.startswith('/'):
                    submit = baseurl + submit

            elif not submit.startswith('http'):
                parts = urlpath.split('/')[0:-1]
                submit = '/'.join(parts) + '/' + submit

            method = form.get('method') or 'POST'

            formstr = "\n- ".join([f"{k} = {valmap[k]}" for k in valmap])
            print(f"Submit {method} to: {submit} with form values:\n\n- {formstr}")

            try:
                if method.upper() == 'POST':
                    resp = requests.post(submit, params=valmap)
                else:
                    resp = requests.get(submit, params=valmap)

                print(f"Unsubscribe form result: {resp.status_code}")
                return resp.status_code == 200
            except Exception as e:
                print(f"Failed to send HTTP request for unsubscribe: {str(e)}")

            return False
    else:
        return True



import re
import imaplib
import mimetypes

from io import BytesIO

from email import policy
from email.parser import BytesParser

LIST_PATTERN = re.compile(r'\((?P<flags>.*?)\) "(?P<delimiter>.*)" (?P<name>.*)')

def open_connection(config):
    print(f"Opening connection to: {config.server} for user: {config.user}")
    connection = imaplib.IMAP4_SSL(config.server)
    connection.login(config.user, config.password)

    return connection

def check_status(connection, config):
    for mname in config.mailboxes:
        check_status_single(connection, mname)

def check_status_single(connection, mname):
    print(f"Checking status on mailbox: {mname}")
    status_typ,status = connection.status(f"\"{mname}\"",'(MESSAGES RECENT UNSEEN)')
    print(f"Status: {status[0].decode('utf-8')[len(mname)+2: -1]}")

def list_mailboxes(connection):
    typ,data = connection.list()
    print(f"Response code: {typ}")
    for line in data:
        match = LIST_PATTERN.match(line.decode('utf-8'))
        flags,delimiter,mname = match.groups()
        mname = mname.strip('"')

        check_status_single(connection, mname)

def mark_read(connection, mbox_messages):
    for mbox in mbox_messages:
        for mid in mbox_messages[mbox]:
            connection.select(f"\"{mbox}\"")

            print(f"Marking message {mid} as read, in mailbox: {mbox}")
            connection.store(mid, '+FLAGS', '(\Seen \Answered)')

def for_message_html(connection, config, fn, max_count=None):
    results = {}
    for mbox in config.mailboxes:
        connection.select(f"\"{mbox}\"", readonly=True)
        stat, raw_id_list = connection.search(None, 'UNSEEN')
        print(f"Raw id list: '{raw_id_list}'")
        msg_ids = list(reversed(list(filter(None, raw_id_list[0].decode('utf-8').split(' ')))))
        if max_count is not None and len(msg_ids) > max_count:
            msg_ids = msg_ids[0:max_count]

        print(f"{stat}: Got {len(msg_ids)} NEW message IDs: {msg_ids}")

        for mid in msg_ids:
            print(f"\nFetching message: {mid}")
            stat, data = connection.fetch(mid, '(RFC822)')
            print(stat)

            if 'OK' != stat:
                print(f"Failed to fetch email {mid}: {stat} {data}")
                continue

            msgdata = None
            for d in data:
                if not isinstance(d[1], int):
                    msgdata = d[1]
                    break

            if msgdata is None:
                print("Can't find valid email message data. Skipping")
                continue

            msg = BytesParser(policy=policy.default).parse(BytesIO(msgdata))

            print(f"Subject: \"{msg['subject']}\"\nFrom: {msg['from']}\nis multipart? {msg.is_multipart()}")

            html = msg.get_body(preferencelist=('html', 'plain')).get_content()
            result = fn(html)
            if result is not None:
                msgs = results.get(result)
                if msgs is None:
                    msgs = {}
                    results[result] = msgs

                box_msgs = msgs.get(mbox)

                if box_msgs is None:
                    box_msgs = []
                    msgs[mbox] = box_msgs

                box_msgs.append(mid)

    return results

def parse_unsub(html):
    regexes = [
        re.compile(r"[Uu]nsubscribe[^<]+<a.+(http[^\"']+)[^>]+>", re.DOTALL),
        re.compile(r"(http[^>'\"]+[Uu]nsubscribe[^\"']+)[^>]+>", re.DOTALL),
        re.compile(r"(http[^\"']+)[^>]+>.*?[Uu]nsubscribe", re.DOTALL),
    ]

    url = None
    for r in regexes:
        print(f"Searching for: {r}")
        smallest = None
        link = None
        for match in re.finditer(r, html):
            if smallest is None or len(smallest) > len(match.group(0)):
                smallest = match.group(0)
                link = match.group(1)

        if link is not None:
            url = link
            break

    print(f"Unsubscribe URL:\n\n    {url}")
    return url

def get_unsubscribe_urls(connection, config, max_count=None):
    return for_message_html(connection, config, parse_unsub, max_count)

def get_text(msg):
    if msg.is_multipart():
        return get_text(msg.get_payload(0))

    return msg.get_payload(None, True);


#!/usr/bin/env python3

from setuptools import setup, find_packages
import sys

setup(
    zip_safe=True,
    name='unsubmail',
    version='1.0.0',
    long_description="Application for processing unsubscribe links in IMAP mail",
    classifiers=[
      "Development Status :: 3 - Alpha",
      "Intended Audience :: Developers",
      "License :: OSI Approved :: GNU General Public License (GPL)",
      "Programming Language :: Python :: 3",
    ],
    keywords='imap spam',
    author='John Casey',
    author_email='jdcasey@commonjava.org',
    url='https://github.com/jdcasey/unsubmail',
    license='GPLv3+',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    install_requires=[
      "ruamel.yaml",
      "click",
      "requests",
      "beautifulsoup4"
    ],
    include_package_data=True,
    test_suite="tests",
    entry_points={
      'console_scripts': [
        'unsubmail = unsubmail:unsubscribe'
      ],
    }
)




import click
import unsubmail.config as config
import unsubmail.imap as imap
import unsubmail.http as http

@click.command()
@click.option('--config-file', '-c', help='Alternative config YAML')
@click.option('--max-count', '-C', help='Maximum messages to process')
def unsubscribe(config_file=None, max_count=None):
    cfg = config.load(config_file)
    with imap.open_connection(cfg) as conn:
        # imap.list_mailboxes(conn)
        imap.check_status(conn, cfg)

        max_to_process = None
        if max_count is not None:
            max_to_process = int(max_count)

        urls = imap.get_unsubscribe_urls(conn, cfg, max_to_process)
        # url_listing = "\n- ".join(urls)
        print(f"\n\nFound {len(urls)} unsubscribe URLs")

        for url in urls:
            messages_containing = urls[url]
            print(f"Processing...\n\nURL: {url}\nFor Messages: {messages_containing}\n\n")

            if http.process_unsub(url) is True:
                print(f"Unsubscribe apparently successful for: {messages_containing}")
                imap.mark_read(conn, messages_containing)
            else:
                print(f"Unsubscribe failed for: {messages_containing}")


